/*
 * Copyright CERN 2013
 * Author: Federico Vaga <federico.vaga@gmail.com>
 */

#ifndef ADC_LIB_H_
#define ADC_LIB_H_

#ifdef __cplusplus
#pragma GCC diagnostic ignored "-Wwrite-strings"
extern "C" {
#endif

#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>

/* Error codes start from 1024 to void conflicting with libc codes */
#define __ADC_ERRNO_START 1024
#define ADC_ENOP		1024
#define ADC_ENOCAP		1025
#define ADC_ENOCFG		1026
#define ADC_ENOGET		1027
#define ADC_ENOSET		1028
#define ADC_ENOCHAN		1029
#define ADC_ENOMASK		1030
#define ADC_EDISABLED		1031
#define ADC_EROUTE		1032
#define ADC_ENOP_SWTRG		1033
#define ADC_ENOP_OFFCLR	1034
#define ADC_ENOP_OFFCLRHW	1035
#define ADC_ENOP_OFFCLRSW	1036
#define ADC_OFF_AC_RESTORE_S	1037
#define ADC_OFF_AC_RESTORE_R	1038
#define ADC_ENOFLAGS		1039

/**
 * Opaque type. any instance of this should be used as token
 * to use the API
 */
struct adc_dev;

extern const char *adc_board_name[];

/**
 * Enumerate all supported boards
 */
enum adc_supported_boards {
	FMCADC_100MS_4CH_14BIT = 0, /**< It identifies the FMC ADC 100M 14Bit
				       4Channel */
	ADC_ZIOFAKE, /**< It identifies a generic board that supports the ZIO
			framework */
	ADC_GENERICFAKE, /**< It identifies a generic board for testing
			    purpose */
	__ADC_SUPPORTED_BOARDS_LAST_INDEX, /**< It represents the the last
					      index of enum. It can be useful
					      for some sort of automation */
};


/**
 * The buffer hosts data and metadata, plus informative fields
 */
struct adc_buffer {
	void *data; /**< container for all the samples */
	void *metadata; /**< description of the acquisition */
	int samplesize; /**< size of a single sample in Byte (a sample for
			   each channel, e.g 4 chan 14bit sample => samplesize = 8 Byte) */
	int nsamples; /**< number of samples in the buffer */
	struct adc_dev *dev; /**< the device used for the acquisition */
	void *mapaddr; /**< mmap address */
	unsigned long maplen; /**<  mmap length */
	unsigned long flags; /**< internal to the library */
	void *priv; /**< library private date */
};


/**
 * Timestamp descriptor. This is should be big enough to store *any*
 * time stamp.
 */
struct adc_timestamp {
	uint64_t secs; /**< number of seconds */
	uint64_t ticks; /**< number of coarse sub-seconds unit */
	uint64_t bins; /**< high resolution sub-seconds unit */
};


/**
 * It describes the possible polarity values
 */
enum adc_trigger_polarity {
	ADC_TRG_POL_POS = 0, /**< positive edge/slope */
	ADC_TRG_POL_NEG, /**< negative edge/slope */
};

/**
 * It describes the possible configuration option for an external trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_ext {
	ADC_CONF_TRG_EXT_ENABLE, /**< It enable/disable the trigger */
	ADC_CONF_TRG_EXT_POLARITY, /**< It is used to apply the polarity to the
				      external triggers */
	ADC_CONF_TRG_EXT_DELAY, /**< acquisition delay after trigger */
	__ADC_CONF_TRG_EXT_LAST_INDEX, /**< It represents the the last index of
					  this enum. It can be useful for some
					  sort of automation */
};


/**
 * It describes the possible configuration option for a threshold trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_thr {
	ADC_CONF_TRG_THR_ENABLE, /**< It enable (1) or disable (0) the trigger */
	ADC_CONF_TRG_THR_POLARITY, /**< It is used to apply the polarity to the
				      threshold triggers */
	ADC_CONF_TRG_THR_DELAY, /**< It defines the acquisition delay after
				   trigger */
	ADC_CONF_TRG_THR_THRESHOLD, /**< The threshold value that triggers
				       the acquisition */
	ADC_CONF_TRG_THR_HYSTERESIS, /**< If defines the hysteresis associated
					to the threshold value */
	__ADC_CONF_TRG_THR_LAST_INDEX, /**< It represents the the last index
					  of this enum. It can be useful for
					  some sort of automation */
};


/**
 * It describes the possible configuration option for a time trigger
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_trigger_tim {
	ADC_CONF_TRG_TIM_ENABLE, /**< It enable (1) or disable (0) the trigger */
	ADC_CONF_TRG_TIM_SECONDS, /**< When to fire: seconds */
	ADC_CONF_TRG_TIM_NANO_SECONDS, /**< When to fire: nanoseconds */
	__ADC_CONF_TRG_TIM_LAST_INDEX, /**< It represents the the last index
					  of this enum. It can be useful for
					  some sort of automation */
};


/**
 * This represents the list of configuration options for the acquisition
 * when using ``conf->type = ADC_CONF_TYPE_ACQ``
 */
enum adc_configuration_acquisition {
	ADC_CONF_ACQ_N_SHOTS = 0, /**< umber of shots during the acquisition */
	ADC_CONF_ACQ_POST_SAMP, /**< number of samples after the trigger */
	ADC_CONF_ACQ_PRE_SAMP, /**<number of sample before the trigger  */
	ADC_CONF_ACQ_UNDERSAMPLE, /**< Acquisition undersample factor */
	ADC_CONF_ACQ_FREQ_HZ, /**< Acquisition frequency */
	ADC_CONF_ACQ_N_BITS, /**< Number of bit per sample */
	__ADC_CONF_ACQ_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * This represents the list of configuration options for the channels
 * when using ``conf->type = ADC_CONF_TYPE_CHN``
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_channel {
	ADC_CONF_CHN_RANGE = 0, /**< Volt range */
	ADC_CONF_CHN_TERMINATION, /**< channel Termination */
	ADC_CONF_CHN_OFFSET, /**< signal offset */
	ADC_CONF_CHN_SATURATION, /**< saturation value */
	ADC_CONF_CHN_GAIN, /**< signal gain */
	__ADC_CONF_CHN_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * This represents the list of configuration options for the board status
 * when using ``conf->type = ADC_CONF_TYPE_BRD``
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_board {
	ADC_CONF_BRD_STATUS = 0, /**< Board status */
	ADC_CONF_BRD_MAX_FREQ_HZ, /**< Maximum frequency in HZ */
	ADC_CONF_BRD_MIN_FREQ_HZ, /**< Minimum frequency in HZ */
	ADC_CONF_BRD_STATE_MACHINE_STATUS, /**< Acquisition state machine status */
	ADC_CONF_BRD_N_CHAN, /**< Number of acquisition channel */
	ADC_CONF_BRD_N_TRG_EXT, /**< Number of external triggers */
	ADC_CONF_BRD_N_TRG_THR, /**< Number of threshold triggers */
	ADC_CONF_BRD_N_TRG_TIM, /**< Number of timer triggers */
	/* TODO seconds, ticks and bins are bit -> High - low */
	ADC_CONF_UTC_TIMING_BASE_S,  /**< Board internal time: seconds */
	ADC_CONF_UTC_TIMING_BASE_T,  /**< Board internal time: coarse
					sub-seconds */
	ADC_CONF_UTC_TIMING_BASE_B,  /**< Board internal time: fine
					sub-seconds*/
	__ADC_CONF_BRD_ATTRIBUTE_LAST_INDEX, /**< It represents the the last
						index of this enum. It can be
						useful for some sort of
						automation */
};


/**
 * It describes the possible configuration types
 * If not specified, the meaning of a configuration option depends on
 * the board in use
 *
 * @todo we should have a generic definition and then the device-specific
 *       code dealing with the conversion
 */
enum adc_configuration_type {
	ADC_CONF_TYPE_BRD = 0, /**< Configuration for the board */
	ADC_CONF_TYPE_CHN, /**< Configuration for an acquisition channels */
	ADC_CONF_TYPE_ACQ, /**< Configuration for the acquisition */
	ADC_CONF_TYPE_CUS, /**< Custom configuration for board-specific
			      options for any of the possible domain: board,
			      channel, triggers. */
	ADC_CONF_TYPE_TRG_EXT, /**< Configuration for external triggers */
	ADC_CONF_TYPE_TRG_THR, /**< Configuration for threshold triggers */
	ADC_CONF_TYPE_TRG_TIM, /**< Configuration for time triggers */
	__ADC_CONF_TYPE_LAST_INDEX, /**< It represents the the last index of
				       this enum. It can be useful for some
				       sort of automation */
};


#define __ADC_CONF_LEN 64 /* number of allocated items in each structure */

/**
 * It enumerates all possible flags for adc_conf.flags
 */
enum adc_conf_flags {
	ADC_CONF_F_ERROR=(1ULL << 0), /**< configuration failed */
};

/**
 * ADC configuration descriptor.
 */
struct adc_conf {
	enum adc_configuration_type type; /**< configuration type */
	uint32_t dev_type; /**< device type */
	uint32_t route_to; /**< internal route to a particular sub-device
			      (e.g. a channel) */
	uint32_t flags; /**< flags from adc_conf_flags */
	uint64_t mask; /**< capabilities mask, 1 when the correspondent
			  value in ``value[]`` is valid */
	uint32_t value[__ADC_CONF_LEN]; /**< array of configuration value.
					   The position in the array correspond
					   to a specific capability */
};


#define ADC_F_USERMASK	0xffff0000 /**< Flag mask -- low-bits are used
				      by lib-int.h */
#define ADC_F_FLUSH	0x00010000 /**< Flag used to flush the buffer
				      (used by adc_open). It is equivalent of
				      adc_acq_flush(adc, ADC_FLUSH_F_ANY) */
#define ADC_F_VERBOSE	0x00020000 /**< Flag used to verbose on stdout/stderr
				      (usable by any function)*/
#define ADC_F_FIXUP	0x00400000 /**< Flag used to fixup a buffer when
				      filling it (usable by adc_fill_buffer) */

/**
 * Enumeration of all possible flags to driver the auto-clear offset
 */
enum adc_offset_auto_clear_flags {
	ADC_OFFSET_AC_F_MANUAL=0x00000001, /**< the signal will be acquired with
					      the last configuration set */
	ADC_OFFSET_AC_F_RESTORE=0x00000002, /**< restore previous
					       configuration when done. It
					       does not have any effect when
					       MANUAL is active*/
	ADC_OFFSET_AC_F_SOFTWARE=0x00000004, /**< use software mechanism,
						it implies manual
						configuration */
	ADC_OFFSET_AC_F_ZERO=0x00000008, /**< it sets the zero-offset to zero;
					    which means that it removes any
					    previously set zero-offset. */
	__ADC_OFFSET_AC_F_MASK=(ADC_OFFSET_AC_F_MANUAL |
				ADC_OFFSET_AC_F_RESTORE |
				ADC_OFFSET_AC_F_SOFTWARE |
				ADC_OFFSET_AC_F_ZERO), /**< used internally */
};

/**
 * Enumeration of all possible flags to drive flushing
 */
enum adc_flush_flags {
	ADC_FLUSH_F_ANY = 0x0,
	ADC_FLUSH_F_SW = 0x1,
	ADC_FLUSH_F_HW = 0x2,
	ADC_FLUSH_F_ALL = ADC_FLUSH_F_SW | ADC_FLUSH_F_HW,
};

/**
 * @defgroup dev Basic
 * Basic library functions
 * @{
 */
extern void adc_print_version();
extern int adc_init(void);
extern void adc_exit(void);
extern char *adc_strerror(int errnum);
extern char *adc_get_driver_type(struct adc_dev *dev);
extern struct adc_dev *adc_open(char *name, unsigned int dev_id,
				      unsigned long totalsamples,
				      unsigned int nbuffer,
				      unsigned long flags);
extern struct adc_dev *adc_open_by_lun(char *name, int lun,
					     unsigned long totalsamples,
					     unsigned int nbuffer,
					     unsigned long flags);
extern int adc_close(struct adc_dev *dev);
extern int adc_trigger_fire(struct adc_dev *dev);
extern int adc_has_trigger_fire(struct adc_dev *dev);
extern int adc_offset_auto_clear(struct adc_dev *dev,
				 unsigned long flags);
/**@}*/


/**
 * @defgroup acq Acquisition
 * Acquisition functions
 * @{
 */
extern int adc_acq_start(struct adc_dev *dev, unsigned int flags,
			    struct timeval *timeout);
extern int adc_acq_poll(struct adc_dev *dev, unsigned int flags,
			    struct timeval *timeout);
extern int adc_acq_stop(struct adc_dev *dev, unsigned int flags);
extern int adc_acq_flush(struct adc_dev *dev, unsigned int flags);
/**@}*/

/**
 * @defgroup conf Configuration
 * Device configuration functions
 * @{
 */
extern void adc_set_conf_mask(struct adc_conf *conf, unsigned int conf_index);
extern void adc_set_conf(struct adc_conf *conf, unsigned int conf_index,
			 uint32_t val);
extern int adc_get_conf(struct adc_conf *conf, unsigned int conf_index,
			uint32_t *val);
extern void adc_set_conf_mask_all(struct adc_conf *conf, struct adc_dev *dev);
/**@}*/

extern int adc_reset_conf(struct adc_dev *dev, unsigned int flags,
			       struct adc_conf *conf);
extern int adc_apply_config(struct adc_dev *dev, unsigned int flags,
			       struct adc_conf *conf);
extern int adc_apply_config_n(struct adc_dev *dev, unsigned int flags,
			      struct adc_conf *conf, unsigned int n);
extern int adc_retrieve_config(struct adc_dev *dev,
				 struct adc_conf *conf);
extern int adc_retrieve_config_n(struct adc_dev *dev,
				 struct adc_conf *conf,
				 unsigned int n);
extern uint64_t adc_get_capabilities(struct adc_dev *dev,
				     enum adc_configuration_type type);
extern int adc_get_param(struct adc_dev *dev, char *name,
			    char *sptr, int *iptr);
extern int adc_set_param(struct adc_dev *dev, char *name,
			    char *sptr, int *iptr);


/**
 * @defgroup buf Buffer
 * Data management functions
 * @{
 */
extern struct adc_buffer *adc_request_buffer(struct adc_dev *dev,
					     int nsamples,
					     void *(*alloc_fn)(size_t),
					     unsigned int flags);
extern int adc_fill_buffer(struct adc_dev *dev,
			   struct adc_buffer *buf,
			   unsigned int flags,
			   struct timeval *timeout);
extern struct adc_timestamp *adc_tstamp_buffer(struct adc_buffer *buf,
					       struct adc_timestamp *);
extern int adc_release_buffer(struct adc_dev *dev,
			      struct adc_buffer *buf,
			      void (*free_fn)(void *));
extern int adc_buffer_get_sample(struct adc_buffer *buf,
				 unsigned int chan,
				 unsigned int acq_sample,
				 int32_t *value);
extern int adc_buffer_fixup(struct adc_buffer *buf);
/**@}*/

/**
 * @defgroup buf_math Buffer Math
 * Mathematical operations on buffers
 * @{
 */
extern int adc_buffer_math_avg(struct adc_buffer *buf,
			       unsigned int chan,
			       int32_t *avg);
/**@}*/

/* libfmcadc version string */
extern const char * const libadc_version_s;

/* zio version string used during compilation of libadc */
extern const char * const libadc_zio_version_s;

/* fmc-adc-100m14b4cha driver version string used during compilation of libadc */
extern const char * const libadc_fmcadc100m14b4cha_version_s;


#ifdef __cplusplus
}
#endif

#endif /* ADC_LIB_H_ */
