/*
 * The ADC library for the specific card
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <linux/zio-user.h>

#include "adc-lib.h"
#include "adc-lib-int.h"

#define ZIO_SYS_PATH "/sys/bus/zio/devices"

#define ADC_NCHAN 4

static int adc_flush_input(struct __adc_dev_zio *fa)
{
	struct zio_control ctrl;
	struct pollfd p = {
		.fd = fa->fdc,
		.events = POLLIN | POLLERR,
	};

	/* Read the control until one is there; data is discarded by zio */
	while (1) {
		int ret;
		int i = poll(&p, 1, 0);
		if (i < 0)
			return -1;
		if ((p.revents & POLLIN) == 0)
			return 0;
		ret = read(fa->fdc, &ctrl, sizeof(ctrl));
		if (ret != sizeof(ctrl))
			return -1;
	}
}

struct adc_dev *adc_zio_open(const struct adc_board_type *b,
			     unsigned int dev_id,
			     unsigned int cset, int chan,
			     unsigned long totalsamples,
			     unsigned int nbuffer,
			     unsigned long flags)
{
	struct __adc_dev_zio *fa;
	struct stat st;
	char *syspath, *devpath, fname[128];
	int udev_zio_dir = 1;
	int ret;

	/* Check if device exists by looking in sysfs */
	ret = asprintf(&syspath, "%s/%s-%04x", ZIO_SYS_PATH,
		       b->devname, dev_id);
	if (ret < 0)
		goto out_str_alloc_syspath;
	if (stat(syspath, &st))
		goto out_fa_stat; /* ENOENT or equivalent */

	/* ZIO char devices are in /dev/zio or just /dev (older udev) */
	if (stat("/dev/zio", &st) < 0)
		udev_zio_dir = 0;
	ret = asprintf(&devpath, "%s/%s-%04x",
		       (udev_zio_dir ? "/dev/zio" : "/dev"),
		       b->devname, dev_id);
	if (ret < 0)
		goto out_str_alloc_devpath;

	/* Sysfs path exists, so device is there, hopefully */
	fa = calloc(1, sizeof(*fa));
	if (!fa)
		goto out_fa_alloc;
	fa->sysbase = syspath;
	fa->devbase = devpath;
	fa->cset = cset;
	fa->chan = chan;
	fa->dev_id = dev_id;

	/* Open char devices */
	if (chan == ~0) {
		sprintf(fname, "%s-%u-i-ctrl", fa->devbase, fa->cset);
		fa->fdc = open(fname, O_RDONLY);
		sprintf(fname, "%s-%u-i-data", fa->devbase, fa->cset);
		fa->fdd = open(fname, O_RDONLY);
	} else {
		sprintf(fname, "%s-%u-%d-ctrl",
			fa->devbase, fa->cset, fa->chan);
		fa->fdc = open(fname, O_RDONLY);
		sprintf(fname, "%s-%u-%d-data",
			fa->devbase, fa->cset, fa->chan);
		fa->fdd = open(fname, O_RDONLY);
	}


	if (fa->fdc < 0 || fa->fdd < 0)
		goto out_fa_open;

	fa->gid.board = b;

	fa->pagesize = getpagesize();

	/* Support verbose operation (turn user flag into internal flag)*/
	if (flags & ADC_F_VERBOSE || getenv("LIB_ADC_VERBOSE"))
		fa->flags |= ADC_FLAG_VERBOSE;

	if (flags & ADC_F_FLUSH) {
		if (adc_acq_flush((void *)&fa->gid, ADC_FLUSH_F_ANY) < 0)
				goto out_fa_open;
	}

	return (void *) &fa->gid;

out_fa_open:
	if (fa->fdc >= 0)
		close(fa->fdc);
	if (fa->fdd >= 0)
		close(fa->fdd);
	free(fa);
out_fa_alloc:
	free(devpath);
out_str_alloc_devpath:
out_fa_stat:
	free(syspath);
out_str_alloc_syspath:
	return NULL;
}

int adc_zio_close(struct adc_dev *dev)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);

	close(fa->fdc);
	close(fa->fdd);
	free(fa->sysbase);
	free(fa->devbase);
	free(fa);
	return 0;
}

/* poll is used by start, so it's defined first */

int adc_zio_acq_poll(struct adc_dev *dev,
			unsigned int flags, struct timeval *to)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	struct pollfd p;
	int to_ms, ret;

	/* So, first sample and blocking read. Wait.. */
	p.fd = fa->fdc;
	p.events = POLLIN | POLLERR;
	if (!to)
		to_ms = -1;
	else
		to_ms = to->tv_sec * 1000 + (to->tv_usec + 500) / 1000;
	ret = poll(&p, 1, to_ms);
	switch (ret) {
	case 0:
		errno = EAGAIN;
		/* fall through */
	case -1:
		return -1;
	}
	if (p.revents & POLLERR) {
		errno = ADC_EDISABLED;
		return -1;
	}

	return 0;
}

int adc_zio_acq_flush(struct adc_dev *dev, unsigned int flags)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);

	if (!((flags & ADC_FLUSH_F_SW) || (flags == ADC_FLUSH_F_ANY))) {
		errno = ADC_ENOFLAGS;
		return -1;
	}

	return adc_flush_input(fa);
}
