/*
 * The ADC library for the specific card
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <linux/zio-user.h>
#include <fmc-adc-100m14b4cha.h> /* the driver one */
#include "adc-lib-100m14b4cha.h" /* the library one */

#include "adc-lib.h"
#include "adc-lib-int.h"

#define ZIO_SYS_PATH "/sys/bus/zio/devices"

#define ADC_NCHAN 4

#define ADC_CONF_GET 0
#define ADC_CONF_SET 1

#define __ADC_CONF_CHN_OFFSET_ZERO 200


#define ADC_100M_4CH_14BIT_ACQ_MASK (1LL << ADC_CONF_ACQ_N_SHOTS) | \
				    (1LL << ADC_CONF_ACQ_POST_SAMP) | \
				    (1LL << ADC_CONF_ACQ_PRE_SAMP) | \
				    (1LL << ADC_CONF_ACQ_UNDERSAMPLE) | \
				    (1LL << ADC_CONF_ACQ_FREQ_HZ) | \
				    (1LL << ADC_CONF_ACQ_N_BITS)
#define ADC_100M_4CH_14BIT_CHN_MASK (1LL << ADC_CONF_CHN_RANGE) | \
				    (1LL << ADC_CONF_CHN_TERMINATION) | \
				    (1LL << ADC_CONF_CHN_OFFSET) | \
				    (1LL << ADC_CONF_CHN_SATURATION)
#define ADC_100M_4CH_14BIT_BRD_MASK (1LL << ADC_CONF_BRD_STATE_MACHINE_STATUS) | \
				    (1LL << ADC_CONF_BRD_N_CHAN) | \
				    (1LL << ADC_CONF_BRD_N_TRG_EXT) | \
				    (1LL << ADC_CONF_BRD_N_TRG_THR) | \
				    (1LL << ADC_CONF_BRD_N_TRG_TIM) | \
				    (1LL << ADC_CONF_UTC_TIMING_BASE_S) | \
				    (1LL << ADC_CONF_UTC_TIMING_BASE_T)
#define ADC_100M_4CH_14BIT_CUS_MASK (1ULL << ADC_CONF_100M14B4CHA_BUF_TYPE) | \
	(1ULL << ADC_CONF_100M14B4CHA_TRG_SW_EN) | \
	(1ULL << ADC_CONF_100M14B4CHA_ACQ_MSHOT_MAX) | \
	(1ULL << ADC_CONF_100M14B4CHA_BUF_SIZE_KB)
#define ADC_100M_4CH_14BIT_TRG_EXT_MASK (1LL << ADC_CONF_TRG_EXT_ENABLE) | \
					(1LL << ADC_CONF_TRG_EXT_POLARITY) | \
					(1LL << ADC_CONF_TRG_EXT_DELAY)
#define ADC_100M_4CH_14BIT_TRG_THR_MASK (1LL << ADC_CONF_TRG_THR_ENABLE) | \
					(1LL << ADC_CONF_TRG_THR_POLARITY) | \
					(1LL << ADC_CONF_TRG_THR_DELAY) | \
					(1LL << ADC_CONF_TRG_THR_THRESHOLD) | \
					(1LL << ADC_CONF_TRG_THR_HYSTERESIS)
#define ADC_100M_4CH_14BIT_TRG_TIM_MASK (1LL << ADC_CONF_TRG_TIM_ENABLE) | \
					(1LL << ADC_CONF_TRG_TIM_SECONDS) | \
					(1LL << ADC_CONF_TRG_TIM_NANO_SECONDS)


static typeof(adc_get_param) *adc_param[] = {
	[ADC_CONF_GET] = adc_get_param,
	[ADC_CONF_SET] = adc_set_param,
};

/**
 * ADC internal data
 */
struct adc_100m14b4cha {
	/**< keep track of all zero-offset values for each range
	   (100mV 1V 10V) */
	int32_t offset_zero[FA100M14B4C_NCHAN][ADC_CONF_100M14B4CHA_CHN_RANGE_N];
};

static int adc_100m14b4cha_offset_zero_set(struct adc_dev *dev);

/**
 * According to totalsamples and nbuffers it resize the buffer to feet
 * the new need. It fails if the rise is not possible for any reason.
 *
 * @return 0 on success, -1 on error and errno is set appropriately
 */
static int adc_100m14b4cha_buffer_resize(struct adc_dev *dev,
					 unsigned long totalsamples,
					 unsigned int nbuffers)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	unsigned long max = nbuffers * (totalsamples * fa->samplesize);
	enum adc_100m14b4cha_buf_type type;
	char str[34];
	int err, val;

	err = adc_get_param(dev, "cset0/current_buffer", str, NULL);
	if (err)
		return err;

	if (strcmp(str, "kmalloc") == 0) {
		type = ADC_CONF_100M14B4CHA_BUF_KMALLOC;
		strcpy(str, "cset0/chani/buffer/max-buffer-len");
	} else if (strcmp(str, "vmalloc") == 0) {
		type = ADC_CONF_100M14B4CHA_BUF_VMALLOC;
		strcpy(str, "cset0/chani/buffer/max-buffer-kb");
	} else {
		return -1;
	}

	err = adc_get_param(dev, str, NULL, &val);
	if (err)
		return err;

	switch (type) {
	case ADC_CONF_100M14B4CHA_BUF_KMALLOC:
		if (totalsamples * fa->samplesize > 4 * 1024 * 1024) {
			errno = EINVAL;
			return -1;
		}
		if (val >= nbuffers)
			return 0;
		val = nbuffers;
		break;
	case ADC_CONF_100M14B4CHA_BUF_VMALLOC:
		/* resize only when necessary */
		if (val * 1024 >= max)
			return 0;
		val = (max / 1024) + 1;
		break;
	}

	return  adc_set_param(dev, str, NULL, &val);
}


/**
 * @copydoc adc_open
 * @param[in] b the board type description
 *
 * It prepares the device for the the usage. We can use some of the ZIO helpers
 */
static struct adc_dev *adc_100m14b4cha_open(const struct adc_board_type *b,
					    unsigned int dev_id,
					    unsigned long totalsamples,
					    unsigned int nbuffer,
					    unsigned long flags)
{
	struct adc_dev *dev = adc_zio_open(b, dev_id, 0, -1, totalsamples,
					   nbuffer, flags);

	/*
	 * We need to save the page size and samplesize.
	 * Samplesize includes the nchan in the count.
	 */

	struct __adc_dev_zio *fa = to_dev_zio(dev);
	int err, bits;

	if (!dev)
		goto out;

	err = adc_get_param(dev, "cset0/resolution-bits", NULL, &bits);
	if (err)
		return NULL;

	fa->samplesize = FA100M14B4C_NCHAN * ((bits / 8) + !!(bits % 8));

	/* try to size it, do not report errors */
	adc_100m14b4cha_buffer_resize(dev, totalsamples, nbuffer);
	dev = (void *)&fa->gid;

	fa->priv = malloc(sizeof(struct adc_100m14b4cha));
	if (!fa->priv)
		return NULL;

	memset(fa->priv, 0, sizeof(struct adc_100m14b4cha));
	err = adc_100m14b4cha_offset_zero_set(dev);
	if (err)
		goto err;

out:
	return dev;

err:
	free(fa->priv);
	return NULL;
}

/**
 * @copydoc adc_close
 */
static int adc_100m14b4cha_close(struct adc_dev *dev)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);

	free(fa->priv);

	return adc_zio_close(dev);
}

/**
 * @copydoc adc_acq_start
 *
 * It starts the acquisition state machine and then it enables the trigger
 */
static int adc_100m14b4cha_acq_start(struct adc_dev *dev,
				     unsigned int flags,
				     struct timeval *timeout)
{
	int cmd = 1; /* hw command for "start" */
	int tmp = 1; /* enable the trigger */
	int err;

	if (flags & ADC_F_FLUSH) {
		err = adc_acq_flush(dev, ADC_FLUSH_F_ANY);
		if (err)
			return err;
	}

	err = adc_set_param(dev, "cset0/trigger/enable", NULL, &tmp);
	if (err)
		return err;
	err = adc_set_param(dev, "cset0/fsm-command", NULL, &cmd);
	if (err)
		return err;

	if (timeout && timeout->tv_sec == 0 && timeout->tv_usec == 0)
		return 0;

	return adc_acq_poll(dev, 0, timeout);
}


/**
 * @copydoc adc_acq_stop
 *
 * It disables the trigger and then it stops the acquisition state machine
 */
static int adc_100m14b4cha_acq_stop(struct adc_dev *dev, unsigned int flags)
{
	int cmd = 2; /* hw command for "stop" */
	int tmp = 0; /* disable the trigger */
	int err;

	err = adc_set_param(dev, "cset0/trigger/enable", NULL, &tmp);
	if (err)
		return err;
	return adc_set_param(dev, "cset0/fsm-command", NULL, &cmd);
}

/**
 * @copydoc adc_acq_flush
 */
static int adc_100m14b4cha_acq_flush(struct adc_dev *dev, unsigned int flags)
{
	return adc_zio_acq_flush(dev, flags);
}

/**
 * It configures the options related to the board
 * @param[in] adc the ADC device token
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_brd(struct adc_dev *adc,
				      unsigned int index,
				      uint32_t *value,
				      unsigned int direction)
{
	switch (index) {
	case ADC_CONF_BRD_STATE_MACHINE_STATUS:
		if (direction)
			return adc_set_param(adc, "cset0/fsm-state",
					     NULL, (int *)value);
		else
			return adc_get_param(adc, "cset0/fsm-state",
					     NULL, (int *)value);
	case ADC_CONF_BRD_N_CHAN:
	case ADC_CONF_BRD_N_TRG_THR:
		if (direction) {
			errno = ADC_ENOSET;
			return -1;
		} else {
			*value = FA100M14B4C_NCHAN;
			return 0;
		}
	case ADC_CONF_BRD_N_TRG_EXT:
		if (direction) {
			errno = ADC_ENOSET;
			return -1;
		} else {
			*value = 1;
			return 0;
		}
	case ADC_CONF_BRD_N_TRG_TIM:
		if (direction) {
			errno = ADC_ENOSET;
			return -1;
		} else {
			*value = 1;
			return 0;
		}
	case ADC_CONF_UTC_TIMING_BASE_S:
		if (direction) {
			errno = ADC_ENOSET;
			return -1;
		} else {
			return adc_get_param(adc, "cset0/tstamp-base-sl",
					     NULL, (int *)value);
		}
	case ADC_CONF_UTC_TIMING_BASE_T:
		if (direction) {
			errno = ADC_ENOSET;
			return -1;
		} else {
			*value = 1;
			return adc_get_param(adc, "cset0/tstamp-base-t",
					     NULL, (int *)value);
		}
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	return 0;
}


static char *options[] = {
	[ADC_CONF_ACQ_N_SHOTS] = "cset0/trigger/nshots",
	[ADC_CONF_ACQ_POST_SAMP] = "cset0/trigger/post-samples",
	[ADC_CONF_ACQ_PRE_SAMP] = "cset0/trigger/pre-samples",
	[ADC_CONF_ACQ_UNDERSAMPLE] = "cset0/undersample",
	[ADC_CONF_ACQ_FREQ_HZ] = "cset0/sample-frequency",
	[ADC_CONF_ACQ_N_BITS] = "cset0/resolution-bits",
};

#define ADC_100M14B4CHA_MASK_RESIZE ((1 << ADC_CONF_ACQ_N_SHOTS) |   \
				     (1 << ADC_CONF_ACQ_PRE_SAMP) |  \
				     (1 << ADC_CONF_ACQ_POST_SAMP))
static int adc_100m14b4cha_config_acq_buffer_resize(struct adc_dev *adc)
{
	unsigned long totalsamples;
	unsigned int nbuffer;
	int err, val;

	err = adc_get_param(adc, options[ADC_CONF_ACQ_N_SHOTS], NULL, &val);
	if (err)
		return err;
	nbuffer = val;

	err = adc_get_param(adc, options[ADC_CONF_ACQ_PRE_SAMP], NULL, &val);
	if (err)
		return err;
	totalsamples = val;

	err = adc_get_param(adc, options[ADC_CONF_ACQ_POST_SAMP], NULL, &val);
	if (err)
		return err;
	totalsamples += val;

	return adc_100m14b4cha_buffer_resize(adc, totalsamples, nbuffer);
}

/**
 * It configures the options related to the acquisition
 * @param[in] adc the ADC device token
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_acq(struct adc_dev *adc,
				      unsigned int index,
				      uint32_t *value,
				      unsigned int direction)
{
	switch (index) {
	case ADC_CONF_ACQ_N_SHOTS:
	case ADC_CONF_ACQ_POST_SAMP:
	case ADC_CONF_ACQ_PRE_SAMP:
	case ADC_CONF_ACQ_UNDERSAMPLE:
	case ADC_CONF_ACQ_FREQ_HZ:
	case ADC_CONF_ACQ_N_BITS:
		return adc_param[direction](adc, options[index],
					    NULL, (int *)value);
		break;
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
}


#define ADC_100M14B4CHA_MAX_CHN_SRC FA100M14B4C_NCHAN
/**
 * It configures the options related to the channel
 * @param[in] adc the ADC device token
 * @param[in] source which of the channels to configure
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_chn(struct adc_dev *adc,
				      unsigned int source,
				      unsigned int index,
				      uint32_t *value,
				      unsigned int direction)
{
	struct __adc_dev_zio *fa = to_dev_zio(adc);
	char path[1024];
	int err;

	if (source > ADC_100M14B4CHA_MAX_CHN_SRC - 1) {
		errno = ADC_EROUTE;
		return -1;
	}

	switch (index) {
	case ADC_CONF_CHN_RANGE:
		sprintf(path, "cset%u/ch%u-vref", fa->cset, source);
		break;
	case ADC_CONF_CHN_TERMINATION:
		sprintf(path, "cset%u/ch%u-50ohm-term", fa->cset, source);
		break;
	case ADC_CONF_CHN_OFFSET:
		sprintf(path, "cset%u/ch%u-offset", fa->cset, source);
		break;
	case __ADC_CONF_CHN_OFFSET_ZERO:
		sprintf(path, "cset%u/ch%u-offset-zero", fa->cset, source);
		break;
	case ADC_CONF_CHN_SATURATION:
		sprintf(path, "cset%u/ch%u-saturation", fa->cset, source);
		break;
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	err = adc_param[direction](adc, path, NULL, (int *)value);
	if (!err && direction == ADC_CONF_SET &&
	    index == ADC_CONF_CHN_RANGE) {
		/*
		 * Whenever the user changes voltage range we need to adjust
		 * the zero-offset value (if any)
		 */
		return adc_100m14b4cha_offset_zero_set(adc);
	}

	return err;
}

/**
 * Enable or disable a trigger source
 * @param[in] adc the ADC device token
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @param[in] mask bitmask for the corresponding trigger source
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int __config_trg_enable(struct adc_dev *adc,
			       uint32_t *value,
			       unsigned int direction,
			       uint32_t mask)
{
	int err, tmp;

	err = adc_get_param(adc, "cset0/trigger/source", NULL, &tmp);
	if (err)
		return err;

	if (direction) {
		if (*value)
			tmp |= mask;
		else
			tmp &= ~mask;
		err = adc_set_param(adc, "cset0/trigger/source",
				    NULL, &tmp);
	} else {
		*value = !!(tmp & mask);
	}

	return err;
}


#define ADC_100M14B4CHA_MAX_EXT_SRC (1)
/**
 * It configures the options related to the external trigger
 * @param[in] adc the ADC device token
 * @param[in] source which of the trigger to configure
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_trg_ext(struct adc_dev *adc,
					  unsigned int source,
					  unsigned int index,
					  uint32_t *value,
					  unsigned int direction)
{
	int err, tmp;

	if (source > ADC_100M14B4CHA_MAX_EXT_SRC - 1) {
		errno = ADC_EROUTE;
		return -1;
	}

	/*
	 * The source is in range [0, N], but some other MACRO/function works
	 * in range [1, N + 1]. So, you may see some `source + 1` here and
	 * there
	 */
	switch (index) {
	case ADC_CONF_TRG_EXT_ENABLE:
		err = __config_trg_enable(adc, value, direction,
					  FA100M14B4C_TRG_SRC_EXT);
		if (err)
			return err;
		break;
	case ADC_CONF_TRG_EXT_POLARITY:
		err = adc_get_param(adc, "cset0/trigger/polarity",
				    NULL, &tmp);
		if (err)
			return err;

		if (direction) {
			if (*value)
				tmp |= FA100M14B4C_TRG_POL_EXT;
			else
				tmp &= ~FA100M14B4C_TRG_POL_EXT;
			err = adc_set_param(adc, "cset0/trigger/polarity",
					    NULL, &tmp);
		} else {
			*value = !!(tmp & FA100M14B4C_TRG_POL_EXT);
		}
		break;
	case ADC_CONF_TRG_EXT_DELAY:
		return adc_param[direction](adc, "cset0/trigger/ext-delay",
					    NULL, (int *)value);
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	return 0;
}


#define ADC_100M14B4CHA_MAX_THR_SRC FA100M14B4C_NCHAN
/**
 * It configures the options related to the threshold trigger
 * @param[in] adc the ADC device token
 * @param[in] source which of the trigger to configure [0, N]
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_trg_thr(struct adc_dev *adc,
					  unsigned int source,
					  unsigned int index,
					  uint32_t *value,
					  unsigned int direction)
{
	char path[1024];
	int err, tmp;

	if (source > ADC_100M14B4CHA_MAX_THR_SRC - 1) {
		errno = ADC_EROUTE;
		return -1;
	}

	/*
	 * The source is in range [0, N], but some other MACRO/function works
	 * in range [1, N + 1]. So, you may see some `source + 1` here and
	 * there
	 */
	switch (index) {
	case ADC_CONF_TRG_THR_ENABLE:
		err = __config_trg_enable(adc, value, direction,
					  FA100M14B4C_TRG_SRC_CHx(source + 1));
		if (err)
			return err;
		break;
	case ADC_CONF_TRG_THR_POLARITY:
		err = adc_get_param(adc, "cset0/trigger/polarity",
				    NULL, &tmp);
		if (err)
			return err;

		if (direction) {
			if (*value)
				tmp |= FA100M14B4C_TRG_POL_CHx(source + 1);
			else
				tmp &= ~FA100M14B4C_TRG_POL_CHx(source + 1);
			return adc_set_param(adc, "cset0/trigger/polarity",
					     NULL, &tmp);
		} else {
			*value = !!(tmp & FA100M14B4C_TRG_POL_CHx(source + 1));
		}
		break;
	case ADC_CONF_TRG_THR_DELAY:
		sprintf(path, "cset0/trigger/ch%u-delay", source);
		return adc_param[direction](adc, path, NULL, (int *)value);
	case ADC_CONF_TRG_THR_THRESHOLD:
		sprintf(path, "cset0/trigger/ch%u-threshold", source);
		return adc_param[direction](adc, path, NULL, (int *)value);
	case ADC_CONF_TRG_THR_HYSTERESIS:
		sprintf(path, "cset0/trigger/ch%u-hysteresis", source);
		return adc_param[direction](adc, path, NULL, (int *)value);
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	return 0;
}

static uint32_t __nanoseconds_to_ticks(uint32_t nanoseconds)
{
	return nanoseconds / FA100M14B4C_UTC_CLOCK_NS;
}

#define ADC_100M14B4CHA_MAX_TIM_SRC (1)
/**
 * It configures the options related to the time trigger
 * @param[in] adc the ADC device token
 * @param[in] source which of the trigger to configure
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_trg_tim(struct adc_dev *adc,
					  unsigned int source,
					  unsigned int index,
					  uint32_t *value,
					  unsigned int direction)
{
	char path[1024];
	uint32_t tmp;
	int err;

	if (source > ADC_100M14B4CHA_MAX_TIM_SRC - 1) {
		errno = ADC_EROUTE;
		return -1;
	}

	switch (index) {
	case ADC_CONF_TRG_TIM_ENABLE:
		err = __config_trg_enable(adc, value, direction,
					  FA100M14B4C_TRG_SRC_TIM);
		if (err)
			return err;
		break;
	case ADC_CONF_TRG_TIM_SECONDS:
		/* FIXME we need 64bit values to configure it completely */
		tmp = 0;
		sprintf(path, "cset0/trigger/trg-time-su");
		err = adc_param[direction](adc, path, NULL, (int *)&tmp);
		if (err)
			return err;
		sprintf(path, "cset0/trigger/trg-time-sl");
		return adc_param[direction](adc, path, NULL, (int *)value);
	case ADC_CONF_TRG_TIM_NANO_SECONDS:
		tmp = __nanoseconds_to_ticks(*value);
		sprintf(path, "cset0/trigger/trg-time-t");
		return adc_param[direction](adc, path, NULL, (int *)&tmp);
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	return 0;
}

/**
 * It configures the options related only to the fmcadc100m14b4cha
 * @param[in] adc the ADC device token
 * @param[in] index the option index
 * @param[in, out] value the value read or to set according to the direction
 * @param[in] direction 0 to read 1 to set
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_100m14b4cha_config_cus(struct adc_dev *adc,
				      unsigned int index,
				      uint32_t *value,
				      unsigned int direction)
{
	char v_str[16];
	int err, tmp;

	switch (index) {
	case ADC_CONF_100M14B4CHA_BUF_TYPE:
		if (direction)
			strcpy(v_str, *value ? "vmalloc" : "kmalloc");
		err = adc_param[direction](adc, "cset0/current_buffer",
					   v_str, NULL);
		if (err)
			return err;
		if (!direction)
			*value = strcmp(v_str, "kmalloc");
		return 0;
	case ADC_CONF_100M14B4CHA_TRG_SW_EN:
		err = adc_get_param(adc, "cset0/trigger/source", NULL, &tmp);
		if (err)
			return err;

		if (direction) {
			if (*value)
				tmp |= FA100M14B4C_TRG_SRC_SW;
			else
				tmp &= ~FA100M14B4C_TRG_SRC_SW;
			return adc_set_param(adc, "cset0/trigger/source",
					     NULL, &tmp);
		} else {
			*value = !!(tmp & FA100M14B4C_TRG_SRC_SW);
		}
		break;
	case ADC_CONF_100M14B4CHA_ACQ_MSHOT_MAX:
		return adc_param[direction](adc, "cset0/max-sample-mshot",
					    NULL, (int *)value);
	case ADC_CONF_100M14B4CHA_BUF_SIZE_KB:
		return adc_param[direction](adc, "cset0/chani/buffer/max-buffer-kb",
					    NULL, (int *)value);
	default:
		errno = ADC_ENOCAP;
		return -1;
	}
	return 0;
}


static int adc_100m14b4cha_config(struct adc_dev *adc,
				  unsigned int flags,
				  struct adc_conf *conf,
				  unsigned int direction)
{
	int err = 0, i, enabled;

	/* Disabling the trigger before changing configuration */
	if (direction) {
		err = adc_get_param(adc, "cset0/trigger/enable",
				    NULL, &enabled);
		if (err)
			return err;
		if (enabled) {
			enabled = 0;
			err = adc_set_param(adc, "cset0/trigger/enable",
					    NULL, &enabled);
			/* restore the initial value */
			enabled = 1;
		}
	}

	for (i = 0; i < __ADC_CONF_LEN; ++i) {
		if (!(conf->mask & (1ULL << i)))
			continue;

		/* Parameter to configure */
		switch (conf->type) {
		case ADC_CONF_TYPE_CUS:
			err = adc_100m14b4cha_config_cus(adc, i,
							 &conf->value[i],
							 direction);
			break;
		case ADC_CONF_TYPE_BRD:
			err = adc_100m14b4cha_config_brd(adc, i,
							 &conf->value[i],
							 direction);
			break;
		case ADC_CONF_TYPE_ACQ:
			err = adc_100m14b4cha_config_acq(adc, i,
							 &conf->value[i],
							 direction);
			break;
		case ADC_CONF_TYPE_CHN:
			err = adc_100m14b4cha_config_chn(adc,
							 conf->route_to,
							 i,
							 &conf->value[i],
							 direction);
			break;
		case ADC_CONF_TYPE_TRG_EXT:
			err = adc_100m14b4cha_config_trg_ext(adc,
							     conf->route_to,
							     i,
							     &conf->value[i],
							     direction);
			break;
		case ADC_CONF_TYPE_TRG_THR:
			err = adc_100m14b4cha_config_trg_thr(adc,
							     conf->route_to,
							     i,
							     &conf->value[i],
							     direction);
			break;
		case ADC_CONF_TYPE_TRG_TIM:
			err = adc_100m14b4cha_config_trg_tim(adc,
							     conf->route_to,
							     i,
							     &conf->value[i],
							     direction);
			break;
		default:
			/*
			 * Already checked in apply/retrive, but just in case that
			 * code disappear I put it here as well
			 */
			errno = ADC_ENOCFG;
			return -1;
		}
		if (err)
			return err;
	}


	/*
	 * if we changed the acquisition parameters we may need to resize
	 * the buffer
	 */
	if (direction && conf->type == ADC_CONF_TYPE_ACQ &&
	    (conf->mask & ADC_100M14B4CHA_MASK_RESIZE)) {
		err = adc_100m14b4cha_config_acq_buffer_resize(adc);
		if (err)
			return err;
	}

	/* if the trigger was enabled restore it */
	if (direction && enabled)
		err = adc_set_param(adc, "cset0/trigger/enable", NULL, &enabled);
	return err;
}


/**
 * @copydoc adc_apply_config
 */
static int adc_100m14b4cha_apply_config(struct adc_dev *dev,
					unsigned int flags,
					struct adc_conf *conf)
{
	return adc_100m14b4cha_config(dev, flags, conf, ADC_CONF_SET);
}


/**
 * @copydoc adc_retrieve_config
 */
static int adc_100m14b4cha_retrieve_config(struct adc_dev *dev,
					   struct adc_conf *conf)
{
	return adc_100m14b4cha_config(dev, 0, conf, ADC_CONF_GET);
}


/**
 * @copydoc adc_trigger_fire
 */
static int adc_100m14b4cha_trigger_fire(struct adc_dev *dev)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	char pathname[128];
	int fd, ret;

        snprintf(pathname, sizeof(pathname),
		 "/sys/kernel/debug/adc-100m14b-%04x/trigger_software",
		 fa->dev_id);
	fd = open(pathname, O_WRONLY);
	if (fd < 0)
		return -1;
	ret = write(fd, "1", 1);
	close(fd);
	if (ret < 0)
		return -1;
	if (ret == 1)
		return 0;
	errno = EIO;
	return -1;
}

static int adc_100m14b4cha_buffer_get_sample(struct adc_buffer *buf,
					     unsigned int chan,
					     unsigned int acq_sample,
					     int32_t *value)
{
	if (chan >= FA100M14B4C_NCHAN) {
		errno = EINVAL;
		return -1;
	}

	*value = ((int16_t *)buf->data)[acq_sample * FA100M14B4C_NCHAN + chan];

	return 0;
}

struct tmp_cfg_store {
	uint32_t trg_source;
	struct adc_conf chn[FA100M14B4C_NCHAN];
	struct adc_conf acq;
};

static int __cfg_offac_save(struct adc_dev *dev,
			    struct tmp_cfg_store *cfg,
			    unsigned long flags)
{
	int err, i;

	if ((flags & ADC_OFFSET_AC_F_MANUAL) ||
	    !(flags & ADC_OFFSET_AC_F_RESTORE))
		return 0;

	cfg->acq.type = ADC_CONF_TYPE_ACQ;
	cfg->acq.mask = (1ULL << ADC_CONF_ACQ_N_SHOTS) |
			(1ULL << ADC_CONF_ACQ_PRE_SAMP) |
			(1ULL << ADC_CONF_ACQ_POST_SAMP) |
			(1ULL << ADC_CONF_ACQ_UNDERSAMPLE);
	err = adc_retrieve_config(dev, &cfg->acq);
	if (err)
		goto err;

	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		memset(&cfg->chn[i], 0, sizeof(struct adc_conf));
		cfg->chn[i].type = ADC_CONF_TYPE_CHN;
		cfg->chn[i].mask = ADC_100M_4CH_14BIT_CHN_MASK;
	}
	err = adc_retrieve_config_n(dev, cfg->chn, FA100M14B4C_NCHAN);
	if (err)
		goto err;

	err = adc_get_param(dev, "cset0/trigger/source", NULL,
			    (int *)&cfg->trg_source);
	if (err)
		goto err;

	return 0;

err:
	errno = ADC_OFF_AC_RESTORE_S;
	return err;
}

static int __cfg_offac_restore(struct adc_dev *dev,
			       struct tmp_cfg_store *cfg,
			       unsigned long flags)
{
	int err = 0;

	if ((flags & ADC_OFFSET_AC_F_MANUAL) ||
	    !(flags & ADC_OFFSET_AC_F_RESTORE))
		return 0;

	err |= adc_apply_config(dev, 0, &cfg->acq);
	err |= adc_apply_config_n(dev, 0, cfg->chn, FA100M14B4C_NCHAN);
	err |= adc_set_param(dev, "cset0/trigger/source", NULL,
			     (int *)&cfg->trg_source);

	if (err)
		errno = ADC_OFF_AC_RESTORE_R;
	return err;
}


/*
 * Following what the configuration for the offset clear computation
 */
#define CFG_OFFAC_NSAMPLES 100000
#define CFG_OFFAC_N 6
static struct adc_conf cfg_offac[CFG_OFFAC_N] ={
	{
		.type = ADC_CONF_TYPE_ACQ,
		.mask = (1ULL << ADC_CONF_ACQ_N_SHOTS) |
			(1ULL << ADC_CONF_ACQ_PRE_SAMP) |
			(1ULL << ADC_CONF_ACQ_POST_SAMP) |
			(1ULL << ADC_CONF_ACQ_UNDERSAMPLE),
		.value = {
			[ADC_CONF_ACQ_N_SHOTS] = 1,
			[ADC_CONF_ACQ_PRE_SAMP] = 0,
			[ADC_CONF_ACQ_POST_SAMP] = CFG_OFFAC_NSAMPLES,
			[ADC_CONF_ACQ_UNDERSAMPLE] = 0,
		},
	},
	{
		.type = ADC_CONF_TYPE_CUS,
		.mask = (1ULL << ADC_CONF_100M14B4CHA_TRG_SW_EN) |
			(1ULL << ADC_CONF_100M14B4CHA_BUF_SIZE_KB),
		.value = {
			[ADC_CONF_100M14B4CHA_TRG_SW_EN] = 1,
			[ADC_CONF_100M14B4CHA_BUF_SIZE_KB] = 2048,
		},
	},
	{
		.type= ADC_CONF_TYPE_CHN,
		.route_to = 0,
		.mask = (1LL << ADC_CONF_CHN_TERMINATION) |
			(1LL << ADC_CONF_CHN_OFFSET) |
			(1LL << ADC_CONF_CHN_SATURATION),
		.value = {
			[ADC_CONF_CHN_TERMINATION] = 0,
			[ADC_CONF_CHN_OFFSET] = 0,
			[ADC_CONF_CHN_SATURATION] = 0x7FFF,
		},
	},
	{
		.type= ADC_CONF_TYPE_CHN,
		.route_to = 1,
		.mask = (1LL << ADC_CONF_CHN_TERMINATION) |
			(1LL << ADC_CONF_CHN_OFFSET) |
			(1LL << ADC_CONF_CHN_SATURATION),
		.value = {
			[ADC_CONF_CHN_TERMINATION] = 0,
			[ADC_CONF_CHN_OFFSET] = 0,
			[ADC_CONF_CHN_SATURATION] = 0x7FFF,
		},
	},
	{
		.type= ADC_CONF_TYPE_CHN,
		.route_to = 2,
		.mask = (1LL << ADC_CONF_CHN_TERMINATION) |
			(1LL << ADC_CONF_CHN_OFFSET) |
			(1LL << ADC_CONF_CHN_SATURATION),
		.value = {
			[ADC_CONF_CHN_TERMINATION] = 0,
			[ADC_CONF_CHN_OFFSET] = 0,
			[ADC_CONF_CHN_SATURATION] = 0x7FFF,
		},
	},
	{
		.type= ADC_CONF_TYPE_CHN,
		.route_to = 3,
		.mask = (1LL << ADC_CONF_CHN_TERMINATION) |
			(1LL << ADC_CONF_CHN_OFFSET) |
			(1LL << ADC_CONF_CHN_SATURATION),
		.value = {
			[ADC_CONF_CHN_TERMINATION] = 0,
			[ADC_CONF_CHN_OFFSET] = 0,
			[ADC_CONF_CHN_SATURATION] = 0x7FFF,
		},
	}
};

/**
 * It configures the ADC for the offset auto-clear
 * @param[in] dev ADC token
 * @param[in] chan channel number [0, 4]
 * @param[in] flags options
 *
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * pre-sample: 0
 * post-sample: 100000
 * undersample: 0
 * nshots: 1
 * offset: 0
 * range: open-drain
 * sw-trg: enable
 * buf-size: 2MiB
 */
static int __cfg_offac_apply(struct adc_dev *dev,
			     unsigned long flags)
{
	int err, trgsrc = FA100M14B4C_TRG_SRC_SW, i;
	uint32_t offset = 0;

	/* zero-offset needs to be cleard always */
	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		err = adc_100m14b4cha_config_chn(dev, i,
						 __ADC_CONF_CHN_OFFSET_ZERO,
						 &offset, ADC_CONF_SET);
		if (err)
			return err;
	}

	if (flags & ADC_OFFSET_AC_F_MANUAL)
		return 0;

	err = adc_set_param(dev, "cset0/trigger/source", NULL,
			    &trgsrc);
	if (err)
		return err;

	err = adc_apply_config_n(dev, 0 , cfg_offac, CFG_OFFAC_N);
	if (err)
		return err;

	return 0;
}

/**
 * It converts an hardware value to uV
 * @param[in] val value to be converted (hardware format)
 * @param[in] range switch configuration read from the hardware
 *
 * @return uV value
 */
static int32_t __convert_hw_to_uv(int32_t val, unsigned int range)
{
	int32_t result_int, result_frac, range_uV, factor;

	switch (range) {
	case ADC_CONF_100M14B4CHA_CHN_RANGE_10V_CAL:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_10V:
		/* 10V */
		range_uV = 10000000;
		break;
	case ADC_CONF_100M14B4CHA_CHN_RANGE_100mV_CAL:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_100mV:
		/* 100mV */
		range_uV = 100000;
		break;
	case ADC_CONF_100M14B4CHA_CHN_RANGE_1V_CAL:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_1V:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_OPEN_DRAIN:
		/* 1V */
		range_uV = 1000000;
		break;
	default:
		fprintf(stderr, "Invalid switch configuration 0x%x\n", range);
		return 0;
	}

	factor = (range_uV / BIT(14));
	result_int = ((val >> 2) * factor);
	result_frac = ((val & 0x3) * 10) * (factor / 1000);

	return result_int + result_frac;
}

static int __offz_rng_idx_get(enum adc_configuration_100m14b4cha_channel_range range)
{
	int idx = 0;

	switch (range) {
	case ADC_CONF_100M14B4CHA_CHN_RANGE_100mV:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_100mV_CAL:
		idx = 0;
		break;
	case ADC_CONF_100M14B4CHA_CHN_RANGE_1V:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_1V_CAL:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_OPEN_DRAIN:
		idx = 1;
		break;
	case ADC_CONF_100M14B4CHA_CHN_RANGE_10V:
	case ADC_CONF_100M14B4CHA_CHN_RANGE_10V_CAL:
		idx = 2;
		break;
	}

	return idx;
}

/**
 * According to the current voltage range, it sets the zero-offset
 * value
 */
static int adc_100m14b4cha_offset_zero_set(struct adc_dev *dev)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	struct adc_100m14b4cha *adcdata = fa->priv;
	int i, err = 0;

	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		uint32_t offset, range;
		int ret;

		ret = adc_100m14b4cha_config_chn(dev, i,
						 ADC_CONF_CHN_RANGE,
						 &range,
						 ADC_CONF_GET);
		if (ret) {
			err |= ret;
			continue;
		}

		offset = adcdata->offset_zero[i][__offz_rng_idx_get(range)];
		err |= adc_100m14b4cha_config_chn(dev, i,
						  __ADC_CONF_CHN_OFFSET_ZERO,
						  &offset, ADC_CONF_SET);
	}

	return err ? -1 : 0;
}

/**
 * It sets a voltage range on all channels
 * @param[in] dev ADC token
 * @param[in] range voltage range
 * @return 0 on success
 */
static int adc_100m14b4cha_range_set_all(struct adc_dev *dev,
					 enum adc_configuration_100m14b4cha_channel_range range)
{
	int i;

	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		int err = adc_100m14b4cha_config_chn(dev, i,
						     ADC_CONF_CHN_RANGE,
						     &range,
						     ADC_CONF_SET);
		if (err)
			return err;
	}
	return 0;
}

/**
 * @copydoc adc_offset_auto_clear
 *
 * MANUAL MODE (user configuration)
 */
static int adc_100m14b4cha_offac_sw_avg_man(struct adc_dev *dev,
					    unsigned long flags)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	struct adc_100m14b4cha *adcdata = fa->priv;
	int32_t offset[FA100M14B4C_NCHAN];
	uint32_t ranges[FA100M14B4C_NCHAN];
	int err, i;

	memset(ranges, 0, sizeof(ranges));
	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		err = adc_100m14b4cha_config_chn(dev, i,
						 ADC_CONF_CHN_RANGE,
						 &ranges[i],
						 ADC_CONF_GET);
		if (err)
			return err;
	}

	err = __cfg_offac_apply(dev, flags);
	if (err)
		return err;


	err = adc_offset_auto_clear_sw_avg(dev, flags, offset);
	if (err)
		return err;

	/*
	 * Since the Hw does (Vin - Vdac), to compensate a
	 * positive offset we need to apply a positive value
	 * to the DAC. For this reason we do not need to revert
	 * that value: the offset sign measured it is fine
	 */
	for (i = 0; i < FA100M14B4C_NCHAN; ++i) {
		int range = __offz_rng_idx_get(ranges[i]);

		offset[i] = __convert_hw_to_uv(offset[i], ranges[i]);
		adcdata->offset_zero[i][range] = offset[i];
	}

	return 0;
}

/**
 * @copydoc adc_offset_auto_clear
 *
 * AUTOMATIC MODE
 */
static int adc_100m14b4cha_offac_sw_avg_auto(struct adc_dev *dev,
					     unsigned long flags)
{
	struct tmp_cfg_store tmpcfg;
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	struct adc_100m14b4cha *adcdata = fa->priv;
	uint32_t ranges[ADC_CONF_100M14B4CHA_CHN_RANGE_N] = {
		ADC_CONF_100M14B4CHA_CHN_RANGE_100mV_CAL,
		ADC_CONF_100M14B4CHA_CHN_RANGE_1V_CAL,
		ADC_CONF_100M14B4CHA_CHN_RANGE_10V_CAL,
	};
	int i, err = 0, err_rst;

	err = __cfg_offac_save(dev, &tmpcfg, flags);
	if (err)
		return err;

	err = __cfg_offac_apply(dev, flags);
	if (err)
		goto out;

	/* Loop over all ranges and do the zero-offset calibration */
	for (i = 0; i < ADC_CONF_100M14B4CHA_CHN_RANGE_N; ++i) {
		unsigned int range = ranges[i];
		int32_t offset[FA100M14B4C_NCHAN];
		int k;

		err = adc_100m14b4cha_range_set_all(dev, range);
		if (err)
			goto out;

		err = adc_offset_auto_clear_sw_avg(dev, flags, offset);
		if (err)
			return err;

		/*
		 * Since the Hw does (Vin - Vdac), to compensate a
		 * positive offset we need to apply a positive value
		 * to the DAC. For this reason we do not need to revert
		 * that value: the offset sign measured it is fine
		 */
		for (k = 0; k < FA100M14B4C_NCHAN; ++k) {
			offset[k] = __convert_hw_to_uv(offset[k], range);
			adcdata->offset_zero[k][i] = offset[k];
		}
	}

out:
	err_rst = __cfg_offac_restore(dev, &tmpcfg, flags);
	if (err_rst) {
		errno = ADC_OFF_AC_RESTORE_R;
		return err_rst;
	}

	return err;
}

/**
 * @copydoc adc_offset_auto_clear
 */
static int adc_100m14b4cha_offset_auto_clear(struct adc_dev *dev,
					     unsigned long flags)
{
	struct __adc_dev_zio *fa = to_dev_zio(dev);
	struct adc_100m14b4cha *adcdata = fa->priv;
	int err;

	/* this board does not have a hardware mechanism */
	flags |= ADC_OFFSET_AC_F_SOFTWARE;

	if (flags & ADC_OFFSET_AC_F_ZERO) {
		memset(adcdata->offset_zero, 0,
		       sizeof(int32_t) * ADC_CONF_100M14B4CHA_CHN_RANGE_N * FA100M14B4C_NCHAN);
		goto out;
	}

	if (flags & ADC_OFFSET_AC_F_MANUAL)
		err = adc_100m14b4cha_offac_sw_avg_man(dev, flags);
	else
		err = adc_100m14b4cha_offac_sw_avg_auto(dev, flags);
	if (err)
		return err;

out:
	return adc_100m14b4cha_offset_zero_set(dev);
}

static struct adc_operations fa_100ms_4ch_14bit_op = {
	.open =		adc_100m14b4cha_open,
	.close =		adc_100m14b4cha_close,

	.acq_start =		adc_100m14b4cha_acq_start,
	.acq_poll =		adc_zio_acq_poll,
	.acq_stop =		adc_100m14b4cha_acq_stop,
	.acq_flush =		adc_100m14b4cha_acq_flush,

	.apply_config =	adc_100m14b4cha_apply_config,
	.retrieve_config =	adc_100m14b4cha_retrieve_config,

	.get_param =		adc_zio_get_param,
	.set_param =		adc_zio_set_param,

	.request_buffer =	adc_zio_request_buffer,
	.fill_buffer =		adc_zio_fill_buffer,
	.tstamp_buffer =	adc_zio_tstamp_buffer,
	.release_buffer =	adc_zio_release_buffer,
	.trigger_fire =	adc_100m14b4cha_trigger_fire,

	.buffer_get_sample =	adc_100m14b4cha_buffer_get_sample,
	.buffer_fixup =	NULL,
	.offset_auto_clear =	adc_100m14b4cha_offset_auto_clear,
};

struct adc_board_type fmcadc_100ms_4ch_14bit = {
	.name = ADC_NAME_FMCADC100M,	/* for library open() */
	.devname = "adc-100m14b",	/* for device named in /dev/zio */
	.driver_type = "zio",
	.capabilities = {
		[ADC_CONF_TYPE_BRD] = ADC_100M_4CH_14BIT_BRD_MASK,
		[ADC_CONF_TYPE_CHN] = ADC_100M_4CH_14BIT_CHN_MASK,
		[ADC_CONF_TYPE_ACQ] = ADC_100M_4CH_14BIT_ACQ_MASK,
		[ADC_CONF_TYPE_CUS] = ADC_100M_4CH_14BIT_CUS_MASK,
		[ADC_CONF_TYPE_TRG_EXT] = ADC_100M_4CH_14BIT_TRG_EXT_MASK,
		[ADC_CONF_TYPE_TRG_THR] = ADC_100M_4CH_14BIT_TRG_THR_MASK,
		[ADC_CONF_TYPE_TRG_TIM] = ADC_100M_4CH_14BIT_TRG_TIM_MASK,
	},
	.adc_op = &fa_100ms_4ch_14bit_op,
};
