/*
 * Initializing and cleaning up the adc library
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */
#include <errno.h>
#include <string.h>
#define ADC_INLINE
#include "adc-lib.h"
#include "adc-lib-int.h"

/* * * * * * * * * * * * * * * * Utilities * * * * * * * * * * * * * * * * */

/**
 * This the table of the error messages
 */
static struct adc_errors {
	int num; /**< the errno that linked to the message */
	char *str; /**< the message associated to the errno */
} adc_errors[] = {
	{ ADC_ENOP,		"Operation not supported"},
	{ ADC_ENOCAP,	"Capabilities not supported"},
	{ ADC_ENOCFG,	"Configuration type not supported"},
	{ ADC_ENOGET,	"Cannot get capabilities information"},
	{ ADC_ENOSET,	"Cannot set capabilities information"},
	{ ADC_ENOCHAN,	"Invalid channel"},
	{ ADC_ENOMASK,	"Missing configuration mask"},
	{ ADC_EDISABLED,	"Trigger is disabled: I/O aborted"},
	{ ADC_EROUTE,	"Cannot route correctly the configuration"},
	{ADC_ENOP_SWTRG, "Operation not supported: software trigger"},
	{ADC_ENOP_OFFCLR, "Operation not supported: offset auto-clear"},
	{ADC_ENOP_OFFCLRHW, "Operation not supported: offset auto-clear hardware"},
	{ADC_ENOP_OFFCLRSW, "Operation not supported: offset auto-clear software"},
	{ADC_OFF_AC_RESTORE_S, "Offset auto-clear: cannot store configuration"},
	{ADC_OFF_AC_RESTORE_R, "Offset auto-clear: cannot restore configuration"},
	{ADC_ENOFLAGS, "Unsupported flags"},
	{ 0, }
};


/**
 * It returns a string describing an error number
 * @param[in] errnum the error number to describe
 * @return a string representing a message error. The function returns static
 *         storage, so you can keep around the pointer it returns, if needed.
 *         Similarly, there is no concurrency between getting the string and
 *         using it, not even in multithreaded environments (``errno`` itself
 *         is already protected, in this respect).
 */
char *adc_strerror(int errnum)
{
	struct adc_errors *p;

	if (errnum < __ADC_ERRNO_START)
		return strerror(errnum);
	for (p = adc_errors; p->num; p++)
		if (p->num == errnum)
			return p->str;
	return "Unknown error code";
}


/**
 * adc_get_driver_type
 * @param[in] dev device which want to know the driver type
 * @return It returns a string that identifies the device type
 */
char *adc_get_driver_type(struct adc_dev *dev)
{
	struct adc_gid *b = (void *)dev;

	return b->board->driver_type;
}


/**
 * It gets the capabilities mask. At least the `type` must be set in `conf`
 * @param[in] dev the ADC device token
 * @param[in] type the configuration type
 * @return It returns capabilities mask for the given configuration type
 */
uint64_t adc_get_capabilities(struct adc_dev *dev,
			      enum adc_configuration_type type)
{
	struct adc_gid *b = (void *)dev;

	if (type >= __ADC_CONF_TYPE_LAST_INDEX)
		return 0;

	return b->board->capabilities[type];
}

/**
 * It sets the bit mask for the given option
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 */
void adc_set_conf_mask(struct adc_conf *conf, unsigned int conf_index)
{
	conf->mask |= (1LL << conf_index);
}

/**
 * It assigns a configuration item, and its mask
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 * @param[in] val value to set
 */
void adc_set_conf(struct adc_conf *conf, unsigned int conf_index,
		  uint32_t val)
{
	conf->value[conf_index] = val;
	adc_set_conf_mask(conf, conf_index);
}

/**
 * It retrieve a configuration value from the configuration descriptor
 * @param[in] conf configuration structure where apply this operation
 * @param[in] conf_index configuration value to set (see configuration enum)
 * @param[out] val value read
 * @return 0 in success, -1 on error
 */
int adc_get_conf(struct adc_conf *conf, unsigned int conf_index,
		 uint32_t *val)
{
	if (conf->mask & (1LL << conf_index)) {
		*val = conf->value[conf_index];
		return 0;
	} else {
		errno = ADC_ENOMASK;
		return -1;
	}
}

/**
 * It enables the configuration for all the supported field by the given device
 * At least the `type` must be set in `conf`
 * @param[in] conf configuration structure to use
 * @param[in] dev the ADC device token
 */
void adc_set_conf_mask_all(struct adc_conf *conf, struct adc_dev *dev)
{
	conf->mask = adc_get_capabilities(dev, conf->type);
}

/**
 * It computes what are the necessary offsets to apply on channels
 * in order to clear a constant offset.
 * @param[in] dev ADC device token
 * @param[in] flags options @see adc_offset_auto_clear_flags
 * @param[out] offset vector of compensation offsets (one for each channel)
 * @return 0 on success, -1 on error and errno is set appropriately
 *         EINVAL: invalid flags value
 *         ADC_ENOP_SWTRG: when software trigger is missing
 *
 * The configuration is board dependent, so here we just run and apply
 * the compensation offset. The configuration is left to the user.
 *
 * Since this function uses software trigger, the user should disable
 * all trigger sources except the software one.
 *
 * NOTE: offset is a vector, so when you call this function be careful
 * about the size of that parameter, it must big enough to host a value
 * for each board channel.
 */
int adc_offset_auto_clear_sw_avg(struct adc_dev *dev,
				 unsigned long flags,
				 int32_t *offset)
{
	struct adc_conf cfg;
	struct adc_buffer *buf;
	struct timeval tv = {0, 0};
	uint32_t nchan = 0, pre = 0, post = 0;
	int err, err_stop, i;
	unsigned int nsamples;

	if (!(flags & ADC_OFFSET_AC_F_SOFTWARE)) {
		errno = EINVAL;
		return -1;
	}

	if (!adc_has_trigger_fire(dev)) {
		errno = ADC_ENOP_SWTRG;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask(&cfg, ADC_CONF_ACQ_PRE_SAMP);
	adc_set_conf_mask(&cfg, ADC_CONF_ACQ_POST_SAMP);
	err = adc_retrieve_config(dev, &cfg);
	if (err)
		return err;
	adc_get_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, &pre);
	adc_get_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, &post);
	nsamples = pre + post;

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask(&cfg, ADC_CONF_BRD_N_CHAN);
	err = adc_retrieve_config(dev, &cfg);
	if (err)
		return err;
	adc_get_conf(&cfg, ADC_CONF_BRD_N_CHAN, &nchan);

	buf = adc_request_buffer(dev, nsamples, NULL, 0);
	if (!buf)
		return -1;

	err = adc_acq_flush(dev, ADC_FLUSH_F_ANY);
	if (err)
		goto out;

	err = adc_acq_start(dev, 0, &tv);
	if (err)
		goto out;

	err = adc_trigger_fire(dev);
	if (err)
		goto out;

	tv.tv_sec = 10;
	err = adc_fill_buffer(dev, buf, 0, &tv);
	if (err)
		goto out;

	for (i = 0; i < nchan; ++i) {
		err = adc_buffer_math_avg(buf, i, &offset[i]);
		if (err)
			goto out;
	}

out:
	adc_release_buffer(dev, buf, NULL);

	err_stop = adc_acq_stop(dev, 0);
	if (err_stop)
		return err_stop;

	return err;
}
